import re

def prepare_var_template(template):
    encounters = {}
    def update_encounter(name):
        nonlocal encounters
        name = name.group(1)
        encounters[name] = name in encounters
        return '.*' if encounters[name] else rf'(?P<{name}>.*)'
    return re.sub(r'{{ *(.*?) *}}', update_encounter, template)



import sonotoria

def load_test_file(file):
    with open(f'tests/test_files/{file}', 'r', encoding='utf-8') as fdesc:
        return fdesc.read()

def test_extract_single_variable():
    # Given
    template = load_test_file('test_extract_single_variable_template.txt')
    input_ = load_test_file('test_extract_single_variable_input.txt')

    # When
    data = sonotoria.extract(template, input_)

    # Then
    assert data['tailor_name'] == 'Henry'

def test_extract_single_variable_with_space():
    # Given
    template = load_test_file('test_extract_single_variable_template.txt')
    input_ = load_test_file('test_extract_single_variable_with_space_input.txt')

    # When
    data = sonotoria.extract(template, input_)

    # Then
    assert data['tailor_name'] == 'John Do'

def test_extract_two_variables():
    # Given
    template = load_test_file('test_extract_two_variables_template.txt')
    input_ = load_test_file('test_extract_two_variables.txt')

    # When
    data = sonotoria.extract(template, input_)

    # Then
    assert data['tailor_last_name'] == 'Williams'
    assert data['tailor_first_name'] == 'Henry'

def test_extract_six_variables_multiline_string():
    # Given
    template = load_test_file('test_extract_six_variables_template.txt')
    input_ = load_test_file('test_extract_six_variables_input.txt')

    # When
    data = sonotoria.extract(template, input_)

    # Then
    assert data['father'] == 'Eric'
    assert data['mother'] == 'Nathalie'
    assert data['nb_siblings'] == 3
    assert data['daughter'] == 'Atalza'
    assert data['cat'] == 'Ellana'
    assert data['fan'] == 'Cassius'

def test_extract_six_variables_with_float():
    # Given
    template = load_test_file('test_extract_six_variables_with_float_template.txt')
    input_ = load_test_file('test_extract_six_variables_with_float_input.txt')

    # When
    data = sonotoria.extract(template, input_)

    # Then
    assert data['father'] == 'Rich'
    assert data['mother'] == 'Tired'
    assert data['nb_siblings'] == 2.6
    assert data['daughter'] == 'LaTex'
    assert data['cat'] == 'Féline'
    assert data['fan'] == '3.14'

def test_extract_variables_with_misc_types():
    # Given
    template = load_test_file('test_extract_variables_with_misc_types_template.txt')
    input_ = load_test_file('test_extract_variables_with_misc_types_input.txt')

    # When
    data = sonotoria.extract(template, input_)

    # Then
    assert data['age'] == 12
    assert data['truth'] is False
    assert data['truth2'] is True
    assert data['number'] == 'Pi'
    assert data['value'] == 2.45
    assert data['place'] == 'church'
    assert data['n1'] == 'true'
    assert data['n2'] == '2.1'
    assert data['bruh'] is None
    assert data['bruh1'] is None
    assert data['bruh2'] is None

def test_extract_variables_used_twice():
    # Given
    template = load_test_file('test_extract_variables_used_twice_template.txt')
    input_ = load_test_file('test_extract_variables_used_twice_input.txt')

    # When
    data = sonotoria.extract(template, input_)

    # Then
    assert data['var'] == 'test'



from sonotoria import yaml

def get_test_file(name):
    return f'tests/yaml_loading_test_files/{name}.yml'

def test_read_simple_yaml():
    # Given
    yaml_file = get_test_file('simple_yaml')

    # When
    data = yaml.load(yaml_file)

    # Then
    assert data == {
        'this': {
            'is': [
                'a',
                'simple',
                {'yaml': 'file'}
            ]
        }
    }

def test_read_simple_json():
    # Given
    json_file = get_test_file('simple_json')

    # When
    data = yaml.load(json_file)

    # Then
    assert data == {
        'this': {
            'is': [
                'a',
                'simple',
                {'yaml': 'file'}
            ]
        }
    }

def test_read_yaml_with_reused_variable():
    # Given
    yaml_file = get_test_file('one_var_yaml')

    # When
    data = yaml.load(yaml_file)

    # Then
    assert data == {
        'my_var': 'test',
        'this': {
            'is': [
                'a',
                'simple',
                {'yaml': 'file'},
                'this is a test'
            ]
        }
    }

def test_read_yaml_with_two_variables():
    # Given
    yaml_file = get_test_file('two_var_yaml')

    # When
    data = yaml.load(yaml_file)

    # Then
    assert data == {
        'my_var': 'test',
        'my_other_var': 'super test',
        'this': {
            'is': [
                'a',
                'simple',
                {'yaml': 'file'},
                'this is a test',
                'this is a super test'
            ]
        }
    }

def test_read_yaml_with_variable_variable():
    # Given
    yaml_file = get_test_file('var_var_yaml')

    # When
    data = yaml.load(yaml_file)

    # Then
    assert data == {
        'my_var': 'test',
        'my_other_var': 'super test',
        'this': {
            'is': [
                'a',
                'simple',
                {'yaml': 'file'},
                'this is a test',
                'this is a super test'
            ]
        }
    }

def test_read_yaml_with_variable_in_map():
    # Given
    yaml_file = get_test_file('map_var_yaml')

    # When
    data = yaml.load(yaml_file)

    # Then
    assert data == {
        'my_var': {
            'test': 'plop'
        },
        'this': {
            'is': [
                'a',
                'simple',
                {'yaml': 'file'},
                'this is a plop'
            ]
        }
    }

def test_read_yaml_with_raw_part():
    # Given
    yaml_file = get_test_file('raw_in_yaml')

    # When
    data = yaml.load(yaml_file)

    # Then
    assert data == {
        'my_var': {
            'test': '{{'
        },
        'this': {
            'is': [
                'a',
                'simple',
                {'yaml': 'file'},
                'this is a {{'
            ]
        }
    }

def test_read_yaml_with_filter():
    # Given
    yaml_file = get_test_file('filter_in_yaml')
    filters = { 'doublereversed': lambda v: f'{v}{v[::-1]}' }

    # When
    data = yaml.load(yaml_file, filters = filters)

    # Then
    assert data == {
        'my_var': {
            'test': 'plop'
        },
        'this': {
            'is': [
                'a',
                'simple',
                {'yaml': 'file'},
                'this is a ploppolp'
            ]
        }
    }

def test_read_yaml_with_test():
    # Given
    yaml_file = get_test_file('test_in_yaml')
    tests = { 'http': lambda v: v.startswith('http://') }

    # When
    data = yaml.load(yaml_file, tests = tests)

    # Then
    assert data == {
        'my_var': {
            'test': 'http://plop',
            'test2': 'https://plop'
        },
        'res': {
            'test': True,
            'test2': False
        },
        'this': {
            'is': [
                'a',
                'simple',
                {'yaml': 'file'}
            ]
        }
    }

def test_read_yaml_with_type():
    # Given
    yaml_file = get_test_file('type_in_yaml')
    class User: # pylint: disable=no-member,too-few-public-methods
        @property
        def name(self):
            return f'{self.first_name.capitalize()} {self.last_name.capitalize()}'
    types = { 'user': User }

    # When
    data = yaml.load(yaml_file, types = types)

    # Then
    assert data['res'].first_name == 'eric'
    assert data['res'].last_name == 'final'
    assert data['res'].name == 'Eric Final'

def test_read_yaml_with_map_var():
    # Given
    yaml_file = get_test_file('map_as_var')

    # When
    data = yaml.load(yaml_file)

    # Then
    assert data == {
        'my_var': {
            'test': 'http://plop',
            'test2': 'https://plop'
        },
        'res': {
            'test': 'http://plop',
            'test2': 'https://plop'
        },
        'this': {
            'is': [
                'a',
                'simple',
                {'yaml': 'file'}
            ]
        }
    }

def test_read_yaml_with_filtered_map_var():
    # Given
    yaml_file = get_test_file('filtered_map_as_var')
    filters = { 'add_plop_to_values': lambda d: { k: f'{v}_lol' for k, v in d.items() } }

    # When
    data = yaml.load(yaml_file, filters = filters)

    # Then
    assert data == {
        'my_var': {
            'test': 'http://plop',
            'test2': 'https://plop'
        },
        'res': {
            'test': 'http://plop_lol',
            'test2': 'https://plop_lol'
        },
        'this': {
            'is': [
                'a',
                'simple',
                {'yaml': 'file'}
            ]
        }
    }

def test_read_yaml_with_var_from_custom_type():
    # Given
    yaml_file = get_test_file('var_from_custom_type')
    class Figure: # pylint: disable=no-member,too-few-public-methods
        pass
    types = {'figure': Figure}

    # When
    figure = yaml.load(yaml_file, types = types)

    # Then
    assert figure.model == 'Goku'
    assert figure.size == '4cm'
    assert figure.material == 'plastic'
    assert figure.serial_number == 'F4521ATB'
    assert figure.price == '45€'
    assert figure.test == 'Goku'

def test_read_yaml_with_context():
    # Given
    context = { 'myvar': 'myval' }
    yaml_file = get_test_file('use_context')

    # When
    figure = yaml.load(yaml_file, context = context)

    # Then
    assert figure['test'] == 'myval'



import re

LOOP_ENTRY_REGEX = r'{%\s*for\s+(?P<key_name>\w+)(\s*,\s*(?P<value_name>\w+))?\s+in\s+(?P<dict_name>\w+)(?:\\\.items\\\(\\\))?\s+%}'
LOOP_EXIT_REGEX = r'{%\s*endfor\s*%}'

class LoopClosedBeforeOpenedException(Exception):
    def __init__(self):
        super().__init__('Loop was closed before it was opened.')

class LoopNeverClosedException(Exception):
    def __init__(self):
        super().__init__('Loop was opened but never closed.')

def invalid(_):
    raise LoopClosedBeforeOpenedException()


LOOKING_FOR_END = 0
LOOKING_FOR_START = 1
TRANSITIONS = {
    LOOKING_FOR_END: {
        'enter': lambda d: (LOOKING_FOR_END, d+1),
        'exit': lambda d: (LOOKING_FOR_END, d-1) if d > 1 else (LOOKING_FOR_START, 0)
    },
    LOOKING_FOR_START: {
        'enter': lambda _: (LOOKING_FOR_END, 1),
        'exit': invalid
    }
}

class ForLoopStateMachine:
    def __init__(self):
        self._state = (LOOKING_FOR_START, 0)

    def transition(self, event):
        self._state = TRANSITIONS[self._state[0]][event](self._state[1])

    @property
    def state(self):
        return self._state[0]

class ForLoop:
    def __init__(self, begin, end, template):
        self.begin = begin[0]
        self.end = end[2].end()
        self.content_begin = begin[2].end()
        self.content_end = end[0]
        self.content = template[self.content_begin:self.content_end].strip('\n')
        self.vars = [begin[2][1], begin[2][3]] if begin[2][3] else [begin[2][1]]
        self.name = begin[2][4]

    @property
    def id(self): #pylint: disable=invalid-name
        return f'loop_{self.begin}_{self.end}'

    @property
    def is_dict(self):
        return len(self.vars) == 2

    def __repr__(self):
        content = self.content.replace('\n', '\\n')
        return f'{self.name}: [{self.begin} - {self.end}] ({", ".join(self.vars)}); content : "{content}"'

def keep_top_loop_only(ids):
    if ids == []: return [] #pylint: disable=multiple-statements

    kept = []
    machine = ForLoopStateMachine()

    for id_ in ids:
        if machine.state == LOOKING_FOR_START:
            kept.append(id_)
        machine.transition(id_[1])
        if machine.state == LOOKING_FOR_START:
            kept.append(id_)

    if machine.state != LOOKING_FOR_START:
        raise LoopNeverClosedException()
    return kept

def identify(template):
    entries = [(entry.start(), 'enter', entry) for entry in re.finditer(LOOP_ENTRY_REGEX, template)]
    exits = [(entry.start(), 'exit', entry) for entry in re.finditer(LOOP_EXIT_REGEX, template)]
    ids = sorted(
        [*entries, *exits],
        key = lambda id_: id_[0]
    )
    return keep_top_loop_only(ids)

def find_top(template):
    ids = identify(template)
    return [
        ForLoop(ids[i], ids[i+1], template)
        for i in range(0, len(ids), 2)
    ]

def extract_top(template, loops):
    for i in range(len(loops)-1, -1, -1):
        template = template[:loops[i].begin] + f'(?P<{loops[i].id}>.*?)' + template[loops[i].end:]
    return template



import sonotoria

from .test_extract_variable import load_test_file


def test_extract_dict():
    # Given
    template = load_test_file('test_extract_dict_template.txt')
    input_ = load_test_file('test_extract_dict_input.txt')

    # When
    data = sonotoria.extract(template, input_)

    # Then
    assert data['parameters'] == {
        'Plop': 'lol',
        'foo': 'bar',
        'PI': 3.14,
        'true': True
    }

def test_extract_dict_template_multiline():
    # Given
    template = load_test_file('test_extract_dict_multiline_template.txt')
    input_ = load_test_file('test_extract_dict_multiline_input.txt')

    # When
    data = sonotoria.extract(template, input_)

    # Then
    assert data['parameters'] == {
        'Plop': 'lol',
        'foo': 'bar',
        'PI': 3.14,
        'true': True
    }

def test_extract_two_dicts():
    # Given
    template = load_test_file('test_extract_two_dicts_template.txt')
    input_ = load_test_file('test_extract_two_dicts_input.txt')

    # When
    data = sonotoria.extract(template, input_)

    # Then
    assert data['parameters'] == {
        'Plop': 'lol',
        'foo': 'bar',
        'PI': 3.14,
        'true': True
    }
    assert data['parameters2'] == {
        'Plop': 'lol',
        'foo': 'bar',
        'PI': 3.14,
        'true': True
    }



import sonotoria

from .test_extract_variable import load_test_file


def test_extract_cond():
    # Given
    template = load_test_file('test_extract_cond_template.txt')
    input_ = load_test_file('test_extract_cond_input.txt')

    # When
    data = sonotoria.extract(template, input_)

    # Then
    assert data['test'] is True
    assert data['test2'] is False
    assert data['bruh'] == 'val1'
    assert data['bruh2'] is None


def test_extract_rest():
    # Given
    template = load_test_file('test_rest_template.txt')
    input_ = load_test_file('test_rest_input.txt')

    # When
    data = sonotoria.extract(template, input_)

    # Then
    assert data['params']['test'] == 'Looks like a test variable, huh'
    assert data['params']['lol'] == 'This might be a fun variable'
    assert data['params']['plop'] == 'Plop might just be the next best variable name'
    assert data['return'] == 'Pretty much nothing, sadly'
    assert data['rtype'] is None
    assert data['return_given'] is True
    assert data['rtype_given'] is False



import os
from setuptools import setup, find_packages

setup(
    name='sonotoria',
    version=os.environ['CI_COMMIT_BRANCH'],
    description='Templating utilities',
    long_description=open('README.md').read()+'\n\n\n'+open('CHANGELOG.md').read(),
    long_description_content_type='text/markdown',
    url='https://gitlab.com/neomyte/sonotoria',
    author='Emmanuel Pluot',
    author_email='emmanuel.pluot@gmail.com',
    packages=find_packages(),
    install_requires=[
        'pyyaml',
        'jinja2'
    ],
    zip_safe=False,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3 :: Only',
        'Operating System :: Unix',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'License :: OSI Approved :: MIT License',
    ]
)


import jinja2

def template_string(str_, data, filters=None, tests=None):
    # Somehow trailing line feed are removed by jinja2 readding them
    add_trail = '\n' if str_.endswith('\n') else ''

    env = jinja2.Environment()
    for name, filter_ in (filters or {}).items():
        env.filters[name] = filter_
    for name, test in (tests or {}).items():
        env.tests[name] = test
    try:
        return env.from_string(str_).render(**data) + add_trail
    except TypeError: # Yaml Objects
        return env.from_string(str_).render(**data.__dict__) + add_trail
